@@ -2,6 +2,7 @@ import os
 from configparser import ConfigParser
 from io import StringIO
 import json
+import pkg_resources
 
 from worklog.breaks import AutoBreak
 import worklog.constants as wc
@@ -12,6 +13,12 @@ from worklog.utils.logger import configure_logger
 from worklog.dispatcher import dispatch
 
 
+try:
+    __version__ = pkg_resources.get_distribution("dcs-" + __name__).version
+except Exception:
+    __version__ = "unknown"
+
+
 def run() -> None:
     """ Main method """
     logger = configure_logger()
